#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import os
import socket
import subprocess
import sys
import time
import warnings
from distutils.spawn import find_executable
from multiprocessing import cpu_count, current_process

import numpy as np
import pytest
from schwimmbad import MultiPool

from uhammer import (Parameters, continue_sampling, persist_every_n_iterations,
                     sample)
from uhammer._parallel import (check_if_walkers_number_of_cores_fit,
                               check_pool, core_count, runs_on_euler_node,
                               runs_on_single_euler_node)


def test_check_euler(monkeypatch):

    assert not runs_on_euler_node()

    for ip in ("10.205.0.10", "10.205.1.10", "10.205.96.0", "10.205.97.0"):

        monkeypatch.setattr(socket, "gethostbyname", lambda name: ip)
        assert runs_on_euler_node()


def test_runs_on_single_euler_node(monkeypatch):
    assert not runs_on_single_euler_node()
    monkeypatch.setenv("LSB_EFFECTIVE_RSRCREQ", "span[hosts=1]")
    assert runs_on_single_euler_node()


def test_core_count(monkeypatch):
    assert core_count() == cpu_count()
    monkeypatch.setenv("LSB_MAX_NUM_PROCESSORS", "124")
    assert core_count() == 124


def test_warning_for_oversubscription(record_warnings, regtest, monkeypatch):
    monkeypatch.setenv("OMP_NUM_THREADS", "124")
    with record_warnings() as messages:
        check_pool(MultiPool(2), 10)
        assert len(messages) == 1, messages
        print("expect warning about oversubscription:".upper(), file=regtest)
        print(messages[0], file=regtest)


def test_check_if_walkers_number_of_cores_fit(record_warnings, regtest):

    with record_warnings() as messages:
        assert not check_if_walkers_number_of_cores_fit(10, 1)
        assert len(messages) == 1, messages
        print(messages[0], file=regtest)

    with record_warnings() as messages:
        check_if_walkers_number_of_cores_fit(10, 20)
        assert len(messages) == 1, messages
        print("expect warning about less walkers than cores:".upper(), file=regtest)
        print(messages[0], file=regtest)

    print(file=regtest)

    with record_warnings() as messages:
        check_if_walkers_number_of_cores_fit(81, 20)
        assert len(messages) == 1, messages
        print(
            "expectwarning about walkers beig multiples of num workers:".upper(),
            file=regtest,
        )
        print(messages[0], file=regtest)

    print(file=regtest)

    with record_warnings() as messages:
        check_if_walkers_number_of_cores_fit(80, 20)
        assert not messages


class PoolMock(object):
    def __init__(self, size):
        self.size = size


def test_check_pool_1(record_warnings, monkeypatch, regtest):
    class PoolMock(object):
        def __init__(self, size):
            self.size = size

    with record_warnings() as messages:
        monkeypatch.setattr(socket, "gethostbyname", lambda name: "10.205.0.10")
        check_pool(PoolMock(36), 72)
        assert len(messages) == 2
        for warning in messages:
            print(warning, file=regtest)


def test_check_pool_2(record_warnings, monkeypatch, regtest):

    with record_warnings() as messages:

        # setup single node on euler:
        monkeypatch.setattr(socket, "gethostbyname", lambda name: "10.205.0.10")
        monkeypatch.setenv("LSB_EFFECTIVE_RSRCREQ", "span[hosts=1]")

        check_pool(PoolMock(36), 72)
        assert len(messages) == 1
        for warning in messages:
            print(warning, file=regtest)


def test_check_pool_3(record_warnings):

    with record_warnings() as messages:
        check_pool(PoolMock(35), 70)
        assert not messages


def lnprob(p):
    # this function must be outside the test functions to be pickable, as
    # multiprocessing pool fails for non-pickable functions.

    time.sleep(0.02)  # make it slow to increase speedup.
    return -p.a ** 2


def test_parallel_sampling(record_warnings, regtest, tmpdir):
    p = Parameters()
    p.add("a", (-1, 1))

    n_samples = 50
    n_walkers_per_parm = 14

    started = time.time()

    samples_serial, __ = sample(
        lnprob, p, [], n_walkers_per_parm, seed=42, n_samples=n_samples, parallel=False
    )

    needed_serial = time.time() - started

    started = time.time()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)

        samples_parallel, __ = sample(
            lnprob,
            p,
            [],
            n_walkers_per_parm,
            seed=42,
            n_samples=n_samples,
            parallel=True,
            persist=persist_every_n_iterations(1, tmpdir.join("sampler.pkl").strpath),
        )
        needed_parallel = time.time() - started

    assert np.all(samples_serial == samples_parallel), (
        samples_serial,
        samples_parallel,
    )

    # we dont check scalablity for less than four cores as my develpment machine has
    # four. For less than four cores it is enough that the assert above did not fail.
    if cpu_count() >= 4:
        assert needed_parallel < 0.6 * needed_serial, (needed_parallel, needed_serial)


def test_mpi_tasks(record_warnings, regtest, monkeypatch):

    from uhammer._parallel import mpi_tasks  # avoids UnboundLocalError

    assert mpi_tasks() is None

    if "mpi4py" in sys.modules:
        del sys.modules["mpi4py"]

    from uhammer._parallel import mpi_tasks

    assert "mpi4py" not in sys.modules, "this test is not working any more"

    class Fake(object):
        @property
        def MPI(self):
            raise ImportError("import fails on purpose.")

    try:
        sys.modules["mpi4py"] = Fake()
        # simulate mpirun:
        os.environ["MPIEXEC_FAKE"] = "LOOKSLIKEMPI"
        with record_warnings() as messages:
            assert mpi_tasks() == 1
    finally:
        del os.environ["MPIEXEC_FAKE"]
        del sys.modules["mpi4py"]

    assert len(messages) == 1, messages
    print(messages[0], file=regtest)


def test_sample_from_lambda():

    seed = 42

    p = Parameters()
    p.add("a", (-1, 1))

    lnprob = lambda s: -abs(s.a)  # noqa

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        sample(
            lnprob,
            p,
            args=None,
            n_walkers_per_param=10,
            seed=seed,
            n_samples=1000,
            show_progress=False,
            parallel=True,
        )


MPI_EXE = find_executable("mpirun")
try:
    import mpipool  # noqa
except ImportError:
    HAS_MPIPOOL = False
else:
    HAS_MPIPOOL = True


@pytest.mark.skipif(HAS_MPIPOOL is False, reason="mpipool not installed")
@pytest.mark.skipif(MPI_EXE is None, reason="mpirun not found")
@pytest.mark.skipif(sys.version_info.major == 2, reason="fails with python")
def test_mpi(tmpdir):

    output_prefix = tmpdir.join("out").strpath
    output_prefix_continue = tmpdir.join("out_continue").strpath
    persist_path = tmpdir.join("sampler.pkl").strpath

    env = os.environ.copy()
    env.update({"OUTPUT_PREFIX": output_prefix, "PERSIST_PATH": persist_path})

    here = os.path.dirname(os.path.abspath(__file__))
    script = os.path.join(here, "files", "sample_line_fit_parallel.py")
    line = [MPI_EXE, "-n", "4", sys.executable, script]

    p = subprocess.Popen(
        line, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env  # shell=True
    )
    found_error = False
    found_shape = False
    for line in iter(p.stdout.readline, b""):
        line = line.rstrip()
        print(str(line, "utf-8"))
        if line.startswith(b"++++ ERROR"):
            found_error = True
        if line.startswith(b"++++ SAMPLES"):
            found_shape = b" ".join(line.split(b" ")[2:])

    assert not found_error
    assert found_shape == b"(1500, 3)"

    pathes = set(os.path.relpath(p.strpath, tmpdir.strpath) for p in tmpdir.listdir())
    expected = [
        os.path.basename(output_prefix) + "_run_0000_worker_{}.txt".format(i)
        for i in range(1, 4)
    ] + [os.path.basename(persist_path)]

    assert pathes == set(expected)

    for worker_id, p in enumerate(expected[:-1], 1):
        full_p = tmpdir.join(p).strpath
        content = set(open(full_p))
        assert content == set((str(worker_id) + "\n",))

    samples, __ = continue_sampling(
        persist_path,
        10,
        parallel=False,
        output_prefix=output_prefix_continue,
        show_output=False,
    )
    assert samples.shape == (10, 3)
    assert os.path.exists(output_prefix_continue + "_run_0000.txt")


def test_multiprocessing(tmpdir):

    output_prefix = tmpdir.join("out").strpath
    output_prefix_continue = tmpdir.join("out_continue").strpath
    persist_path = tmpdir.join("sampler.pkl").strpath

    sigma = .5

    def gen_data():
        a0 = .5
        b0 = .5
        c0 = 1

        x = np.linspace(-2, 2, 100)
        y_measured = a0 + b0 * x + c0 * x ** 2 + sigma * np.random.randn(*x.shape)
        return x, y_measured

    p = Parameters()
    p.add("a", (0, 1))
    p.add("b", (0, 1))
    p.add("c", (0, 2))

    def lnprob(p, x, y_measured, sigma=sigma):
        print(current_process().rank)
        y = p.a + p.b * x + p.c * x ** 2
        diff = (y - y_measured) / sigma
        return -np.dot(diff.T, diff) / 2

    n_samples = 1500
    n_walkers_per_param = 200

    samples, __ = sample(
        lnprob,
        p,
        args=gen_data(),
        n_walkers_per_param=n_walkers_per_param,
        n_samples=n_samples,
        show_progress=False,
        show_output=False,
        output_prefix=output_prefix,
        parallel=True,
        verbose=True,
        persist=persist_every_n_iterations(1, persist_path),
    )

    pathes = set(os.path.relpath(p.strpath, tmpdir.strpath) for p in tmpdir.listdir())
    expected = [
        os.path.basename(output_prefix) + "_run_0000_worker_{}.txt".format(i)
        for i in range(1, cpu_count())
    ] + [os.path.basename(persist_path)]

    assert pathes == set(expected)

    for worker_id, p in enumerate(expected[:-1], 1):
        full_p = tmpdir.join(p).strpath
        content = set(open(full_p))
        assert content == set((str(worker_id) + "\n",))

    samples, __ = continue_sampling(
        persist_path,
        10,
        parallel=False,
        output_prefix=output_prefix_continue,
        show_output=False,
    )
    assert samples.shape == (10, 3)

    assert os.path.exists(output_prefix_continue + "_run_0000.txt")



import warnings
from contextlib import contextmanager
from functools import partial

import pytest


@pytest.fixture
def record_warnings(monkeypatch):
    def record(messages, warning_exception, *a, **kw):
        text = str(warning_exception)
        if "Use get_result() which forces correct exception handling" in text:
            # this is a warning triggered by bug in pytest.
            return
        messages.append(text)

    @contextmanager
    def manager():
        messages = []
        monkeypatch.setattr(warnings, "warn", partial(record, messages))
        yield messages
        monkeypatch.undo()

    yield manager



import ipaddress
import os
import socket
import sys
import warnings
from multiprocessing import Queue, cpu_count, current_process

from schwimmbad import MultiPool as UMultiPool

from ._utils import is_writable, mute_stderr

try:
    from mpipool import Pool as MPIPool
except ImportError:

    class MPIPool(object):
        def __getattr__(self, name):
            raise ImportError("you must run 'pip install mpipool' first.")


if sys.version_info.major == 2:
    ModuleNotFoundError = None
    unicode = str


def runs_on_euler_node():
    # ip ranges according to https://scicomp.ethz.ch/wiki/Cluster_IP_ranges#Euler
    try:
        ip4_text = ipaddress.IPv4Address(socket.gethostbyname(socket.gethostname()))
        return any(
            ip4_text in ipaddress.IPv4Network(network)
            for network in ("10.205.0.0/19", "10.205.96.0/19")
        )
    except Exception:
        ip4_text = ipaddress.IPv4Address(
            unicode(socket.gethostbyname(socket.gethostname()))
        )
        return any(
            ip4_text in ipaddress.IPv4Network(network)
            for network in (unicode("10.205.0.0/19"), unicode("10.205.96.0/19"))
        )


def runs_on_single_euler_node():
    resources = os.environ.get("LSB_EFFECTIVE_RSRCREQ", "")
    return "span[hosts=1]" in resources


def runs_with_mpi():
    return any(
        n.startswith(prefix)
        for n in os.environ.keys()
        for prefix in ("MPIEXEC_", "OMPI_COMM_WORLD_")
    )


def mpi_size():
    if runs_with_mpi():
        from mpi4py import MPI

        return MPI.COMM_WORLD.size
    return None


def _inject_rank_information(path, rank):
    if "{worker_id" in path:
        path = path.format(worker_id=rank)
    return path


def check_and_fix_path(path, rank):

    if rank is not None:
        path = _inject_rank_information(path, rank)

    path = os.path.abspath(path)
    folder = os.path.dirname(path)

    if folder and not os.path.exists(folder):
        os.makedirs(folder)

    if os.path.exists(path):
        raise IOError("{} already exists".format(path))

    if not is_writable(path):
        raise IOError(
            "can not write to {}. please check if folder {} is writable".format(
                path, folder
            )
        )
    return path


def get_rank():
    if runs_with_mpi():
        from mpi4py import MPI

        return MPI.COMM_WORLD.rank

    return getattr(current_process(), "rank", 0)  # main process has no rank set


def core_count():
    lsb_cores = os.environ.get("LSB_MAX_NUM_PROCESSORS")
    if lsb_cores is None:
        return cpu_count()
    return int(lsb_cores)


def mpi_tasks():
    return _mpi_query("size", default=1)


def _mpi_query(attribute, default=None):

    if not runs_with_mpi():
        return None
    try:
        from mpi4py import MPI
    except ImportError as e:
        if ModuleNotFoundError is not None and isinstance(e, ModuleNotFoundError):
            warnings.warn("mpi4py is not installed.")
            return default
        warnings.warn("mpi4py is installed, but I can not import it: {}.".format(e))
        return default
    return getattr(MPI.COMM_WORLD, attribute)  # pragma: no cover


def mpi_rank():
    return _mpi_query("rank", default=1)


def check_if_walkers_number_of_cores_fit(n_walkers, n_workers):

    if n_workers == 1:
        warnings.warn("only one worker available. switch to serial mode.")
        return False

    if n_walkers < n_workers:
        msg = (
            "the number of walkers {n_walkers} is less than the number of workers "
            "{n_workers}.".format(n_walkers=n_walkers, n_workers=n_workers)
        )
        warnings.warn(msg)

    else:
        if n_walkers % n_workers:
            msg = (
                "uhammer runs with {n_workers} workers for {n_walkers} walkers. "
                "resource usage is best if the number walkers is a multiple of number "
                "of workers.".format(n_walkers=n_walkers, n_workers=n_workers)
            )
            warnings.warn(msg)
    return True


def get_pool_for(n_walkers):

    pool = _get_pool(n_walkers)
    check_pool(pool, n_walkers)
    return pool


def check_parallel(parallel):
    if not parallel and runs_with_mpi():
        warnings.warn(
            "looks like you use mpirun to run your code, but set parallel=False"
        )


def _set_rank(q):
    current_process().rank = q.get()


def _get_pool(n_walkers):

    tasks = mpi_tasks()

    if tasks is not None and tasks >= 2:  # at least one master and one worker
        return MPIPool()  # pragma: no cover

    n_workers = core_count() - 1

    # it does not make sense to use more cores than walkers
    n_cores = min(n_workers, n_walkers)

    # setup queue to assign rank attribut to all worker processes:
    q = Queue(n_cores)
    for rank in range(1, n_cores + 1):
        q.put(rank)

    with mute_stderr():
        # might cause MPI related error message, even without mpirun:
        pool = UMultiPool(n_cores, initializer=_set_rank, initargs=(q,))
        current_process().rank = 0

    return pool


def check_pool(pool, n_walkers):

    omp_threads = os.environ.get("OMP_NUM_THREADS", 1)
    try:
        omp_threads = int(omp_threads)
    except ValueError:
        raise RuntimeError(
            "OMP_NUM_THREADS has value {} which is not an integer".format(omp_threads)
        )
    if omp_threads > 1:
        warnings.warn(
            "you set OMP_NUM_THREADS to {}. this could cause very slow execution "
            "due to oversubscription. to fix this set OMP_NUM_THREADS to 1".format(
                omp_threads
            )
        )

    check_passed = check_if_walkers_number_of_cores_fit(n_walkers, pool.size)
    if not check_passed:
        return None

    if isinstance(pool, MPIPool):
        return  # pragma: no cover

    if runs_on_euler_node():

        if pool.size > 35:
            warnings.warn(
                "on euler best use -n 24 or -n 36 for scalability. "
                "larger numbers don't increase performance."
            )
        if not runs_on_single_euler_node():
            warnings.warn(
                "better run the job with '-R fullnode' to benefit "
                "from multicore parallelisation."
            )



# encoding: utf-8
import sys

import pkg_resources

from ._parallel import runs_with_mpi as _runs_with_mpi
from .lnprob_adapters import *  # noqa
from .parameters import Parameters  # noqa
from .persisting import load_samples  # noqa
from .persisting import persist_every_n_iterations  # noqa
from .persisting import persist_final  # noqa
from .persisting import persist_on_error  # noqa
from .sampler import *  # noqa

if _runs_with_mpi():

    def handler(type_, value, tb):
        import traceback
        from mpi4py import MPI

        traceback.print_tb(tb, file=sys.stdout)
        sys.stdout.flush()

        MPI.COMM_WORLD.Abort(1)

    sys.excepthook = handler

__version__ = pkg_resources.require(__package__)[0].version



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import io
import sys

from uhammer._utils import capture

# Copyright © 2018 Uwe Schmitt <uwe.schmitt@id.ethz.ch>

if sys.version_info.major == 2:
    StringIO = io.BytesIO
else:
    StringIO = io.StringIO


def test_capture_false(regtest, tmpdir):
    stream = StringIO()
    sys.stdout = sys.stderr = stream

    with capture(path=None, show_output=True):
        print(1)
        print(2, file=sys.stderr)

    assert stream.getvalue() == "1\n2\n"

    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


def test_capture_true(regtest, tmpdir):

    stream = StringIO()
    sys.stdout = sys.stderr = stream

    with capture(path=None, show_output=False):
        print(1)
        print(2, file=sys.stderr)

    assert stream.getvalue() == ""

    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


def test_capture_to_file(regtest, tmpdir):

    capture_file_path = tmpdir.join("out.txt").strpath

    with capture(path=capture_file_path, show_output=False):
        print(1)
        print(2, file=sys.stderr)

    assert open(capture_file_path).read() == "1\n2\n"



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import os
from contextlib import contextmanager
from functools import partial

import emcee
import pytest

from uhammer import (Parameters, continue_sampling, load_samples,
                     persist_every_n_iterations, persist_final,
                     persist_on_error, sample)

emcee_version = int(emcee.__version__.split(".")[0])


@pytest.fixture(scope="function")
def sampler_with_persister(tmpdir):
    def dummy(p):

        return -abs(3 * p.a)

    @contextmanager
    def create_sampler(expected_shape_of_samples, persister, lnprob=dummy):

        p = Parameters()
        p.add("a", (-4, 4))

        assert len(tmpdir.listdir()) == 0

        persist_file = tmpdir.join("sampler.bin").strpath
        persist = persister(path=persist_file)
        sample_ = partial(
            sample,
            lnprob,
            p,
            None,
            n_walkers_per_param=20,
            seed=42,
            n_samples=100,
            persist=persist,
            show_output=False,
        )
        yield sample_, persist

        assert os.path.exists(persist_file)

        # TODO:
        # emcee 3 only returns successful samples, wheras
        # emcee 2 seems to return zeros filled in:
        samples = load_samples(persist_file)
        assert samples.shape == expected_shape_of_samples
        os.remove(persist_file)

    return create_sampler


def test_persist_on_error(sampler_with_persister):
    def error_ln_prob(p, count=[0]):

        count[0] += 1
        if count[0] > 40:
            # we do this to bypass the exception handler during the first iteration.  we
            # have 20 walkers so one iteration involves calling this function 20
            # times.
            1 / 0
        return -abs(3 * p.a)

    # emcee 3 makes some extra function calls before actual sampling starts:
    expected_shape = (40, 1) if emcee_version == 2 else (20, 1)

    with sampler_with_persister(expected_shape, persist_on_error, error_ln_prob) as (
        sample_,
        persister,
    ):

        with pytest.raises(ZeroDivisionError):
            sample_()


def test_persist_every_n_iterations(sampler_with_persister):

    with sampler_with_persister((100, 1), partial(persist_every_n_iterations, 1)) as (
        sample_,
        persister,
    ):
        sample_()
    assert persister._iter_track == [1, 2, 3, 4, 5]

    with sampler_with_persister((80, 1), partial(persist_every_n_iterations, 2)) as (
        sample_,
        persister,
    ):
        sample_()

    assert persister._iter_track == [2, 4]


def test_persist_final(sampler_with_persister):

    with sampler_with_persister((100, 1), persist_final) as (sample_, persister):
        sample_()

    assert persister._iter_track == [5]


def test_restore_sampler(sampler_with_persister, regtest):

    regtest.identifier = "emcee_" + str(emcee_version)

    with sampler_with_persister((100, 1), persist_final) as (sample_, persister):

        samples, __ = sample_()
        print(samples[-10:, :], file=regtest)

        samples, __ = continue_sampling(persister.path, 10)
        assert samples.shape == (10, 1)
        print(samples, file=regtest)



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import numpy as np

from ._utils import check_range, is_float
from .parameters import Parameters

# Copyright © 2018 Uwe Schmitt <uwe.schmitt@id.ethz.ch>


def gaussian(cov):
    """Returns logarithmic probability density function for gaussian with mean 0 and
    given covariance `cov`.

    :param cov: Covariance as a single number (real case) or square invertible
                matrix (multivariate case).
    """

    class ComputeProb(object):
        def __init__(self, cov):
            self.icov = None
            self.cov = cov

        def _setup_icov(self, n):
            cov = np.atleast_1d(self.cov)
            if cov.shape == (1,):
                icov = 1.0 / cov * np.eye(n)
            else:
                icov = np.linalg.inv(cov)
            assert icov.shape == (
                n,
                n,
            ), "'cov' must be a square matrix of size {} but has shape {} x {}".format(
                n, icov.shape[0], icov.shape[1]
            )
            self.icov = icov

        def __call__(self, x):
            if self.icov is None:
                self._setup_icov(len(x))
            return -np.dot(x.T, np.dot(self.icov, x)) / 2.0

    return ComputeProb(cov)


def model(forward, probability, input_data, measured_data):
    """
    Creates a function used as `lnprob` argument for ``uhammer.sample`` function..

    :param forward: models forward computation. Takes two arguments: `p`
                    which must be an instance of ``Parameters`` and the given
                    ``input_data``.

    :param probability: A function which computes to probability of difference
                        of the computed forward model and the given
                        ``measured_data``.

    :param input_data: Input data passed as second argument to ``forward``.

    :param measured_data: Measured data. Must be compatible with the return
                          value of ``forward``.
    """

    def likelihood(p, *args):
        diff = forward(p, input_data) - measured_data
        return probability(diff)

    return likelihood


def distribution(prob_density_function, ndim, ranges):
    """
    Helper to sample from a distribution.

    Returns ``lnprob`` function and ``Parameters`` object to be used in sample.

    :param prob_density_function: Function taking vector of length ``ndim``,
                                  must return logarithm of a probability density.

    :param ndim: Dimension of input to ``prob_density_function``.

    :param ranges: Either tuple or list of tuples of numbers to restrict coordinates
                   when sampling.
    """

    tl = (tuple, list)

    assert isinstance(ranges, tl)
    assert len(ranges) > 0

    if len(ranges) == 2:
        check_range(ranges)
        min_, max_ = ranges
        if is_float(min_) and is_float(max_):
            assert min_ < max_, ranges
            ranges = [ranges] * ndim

    for range_ in ranges:
        check_range(range_)

    p = Parameters()
    for i, range_ in enumerate(ranges):
        p.add("x{i}".format(i=i), range_)

    def lnprob(p):
        return prob_density_function(np.array(p.values))

    return lnprob, p



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import functools
import io
import os
import random
import sys
import threading
import time
from contextlib import contextmanager

try:
    from shutil import get_terminal_size
except ImportError:  # pragma: no cover
    from backports.shutil_get_terminal_size import get_terminal_size  # pragma: no cover


is_py_3 = sys.version_info.major == 3


def _console_dimensions():

    try:
        columns, rows = get_terminal_size((80, 25))
    except OSError:
        return 80, 25
    return columns, rows  # pragma: no cover


def _format_time(t):
    if t is None:
        return "??:??:??.?"
    minutes, seconds = divmod(t, 60)
    hours, minutes = divmod(minutes, 60)
    return "{:02.0f}:{:02.0f}:{:04.1f}".format(hours, minutes, seconds)


def _format_time_part(t_passed, t_left):
    time_str_passed = "passed: " + _format_time(t_passed)
    time_str_left = "left: " + _format_time(t_left)
    return time_str_passed + " " + time_str_left + " - "


class _Formatter(object):
    def __init__(self, bar_size, n_steps, started):
        self.bar_size = bar_size
        self.n_steps = n_steps
        self.started = started
        self.time_left = 0

    def format_progress_line(self, symbol, i, tl=None):
        n = self.bar_size * i // self.n_steps
        time_passed = time.time() - self.started

        progress_symbol = "∣"

        if tl is None:
            time_left_estimate = time_passed / (i + 1) * (self.n_steps - i - 1)

            # exponential smoothing, avoid "flattering" for fast sampling:
            self.time_left = 0.8 * self.time_left + 0.2 * time_left_estimate
        else:
            self.time_left = tl
        time_str = _format_time_part(time_passed, self.time_left)
        return "\r{} {}[{}{}]".format(
            symbol, time_str, progress_symbol * n, " " * (self.bar_size - n)
        )


def _write(what):
    # write to systems stdout directly as sys.stdout might be redirected due to
    # capturing
    if is_py_3:
        os.write(1, what.encode("utf-8"))
    else:
        os.write(1, what)  # pragma: no cover


class _UpdateThread(threading.Thread):
    def __init__(self, symbols, formatter, done_symbol, dt=0.07):
        threading.Thread.__init__(self)
        self.symbols = symbols
        self.formatter = formatter
        self.done_symbol = done_symbol
        self.dt = dt

        self.progress_counter = 0

    def run(self):
        self.iter_running = True
        symbol_idx = 0
        try:
            while self.iter_running:
                time.sleep(self.dt)
                symbol = self.symbols[symbol_idx]
                symbol_idx = (symbol_idx + 1) % len(self.symbols)
                progress_line = self.formatter.format_progress_line(
                    symbol, self.progress_counter
                )
                _write(progress_line)
            progress_line = self.formatter.format_progress_line(
                self.done_symbol, self.progress_counter, 0
            )
            _write(progress_line)
            _write("\n")
        except KeyboardInterrupt:
            pass


def with_progressbar(n_steps, iterable, width=None, delta_t=0.05):

    if width is None:
        width, __ = _console_dimensions()

    symbols = random.choice(
        [
            "◰ ◳ ◲ ◱",
            "▁▂▃▄▅▆▇███▇▆▅▄▃▁▁",
            "▏▎▍▌▋▊███▊▋▌▍▎▏",
            "▏▎▍▌▋▊▉███▏",
            "▖▘▝▗",
            "◐ ◓ ◑ ◒ ",
            "◴ ◷ ◶ ◵",
            "⣾⣽⣻⢿⡿⣟⣯⣷",
            "⠁⠂⠄⡀⢀⠠⠐⠈",
            "◢ ◣ ◤ ◥ ",
            ".oOo",
            "⎺⎻⎼⎽⎼⎻⎺",
            "⎽⎼⎻⎺⎺⎽",
        ]
    ).replace(" ", "")

    done_symbol = "✗"

    time_str = _format_time_part(0, None)
    bar_size = min(n_steps, width - 6 - len(time_str))

    started = time.time()
    formatter = _Formatter(bar_size, n_steps, started)

    thread = _UpdateThread(symbols, formatter, done_symbol)
    thread.start()

    try:
        for result in iterable:
            thread.progress_counter += 1
            yield result
            if not thread.is_alive():
                # usually KeyboardInterrupt happened in update thread
                break
    except KeyboardInterrupt:  # pragma: no cover
        thread.iter_running = False  # pragma: no cover
        thread.join()  # pragma: no cover
        raise  # pragma: no cover

    finally:
        # stop thread:
        thread.iter_running = False
        thread.join()


def is_float(x):
    try:
        float(x)
    except ValueError:
        return False
    else:
        return True


def check_range(range_):
    if not isinstance(range_, (list, tuple)) or len(range_) != 2:
        raise ValueError(
            "range_ must be list or tuple of length 2, got {!r}".format(range_)
        )

    min_, max_ = range_

    if not is_float(min_) or not is_float(max_):
        raise ValueError(
            "range_ must be a tuple of 2 float values, got {!r}".format(range_)
        )
    if not min_ < max_:
        raise ValueError(
            "min value {} is not smaller than max value {}".format(min_, max_)
        )


def is_writable(path):

    if os.path.exists(path):
        return os.access(path, os.W_OK)
    try:
        open(path, "w").close()
    except IOError as e:
        print("ERROR", e)
        return False
    try:
        os.unlink(path)
    except FileNotFoundError as e:  # possible race condition when ran in parallel
        print("ERROR", e)
        pass
    return True


class Tee(io.TextIOBase):
    def __init__(self, file_stream, console_stream):
        self.file_stream = file_stream
        self.console_stream = console_stream

    def write(self, *args):
        if self.file_stream is not None:
            self.file_stream.write(*args)
        if self.console_stream is not None:
            self.console_stream.write(*args)

    def close(self):
        pass


@contextmanager
def capture(path, show_output):

    if path is not None:
        fh = open(path, "a")
    else:
        fh = None

    stdout_stream = Tee(fh, sys.stdout if show_output else None)
    stderr_stream = Tee(fh, sys.stderr if show_output else None)

    old_stdout = sys.stdout
    old_stderr = sys.stderr
    try:
        sys.stdout = stdout_stream
        sys.stderr = stderr_stream
        yield

    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        if fh is not None:
            fh.close()


def os_write_works():
    try:
        os.write(1, b"")
    except IOError:
        return False
    else:
        return True


# from https://stackoverflow.com/questions/6796492/temporarily-redirect-stdout-stderr:


@contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:

    with stdchannel_redirected(sys.stderr, os.devnull):
        ...
    """

    oldstdchannel = None
    dest_file = None
    dest_file = open(dest_filename, "w")
    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        os.dup2(dest_file.fileno(), stdchannel.fileno())
        yield
    except IOError:
        # sth with stream handling went wrong, happens e.g. in Jupyter notebooks.
        # In this case we don't redirect:
        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


mute_stderr = functools.partial(stdchannel_redirected, sys.stderr, os.devnull)



# encoding: utf-8
from __future__ import absolute_import, division, print_function

import functools
import glob
import itertools
import os
import random
import sys
import warnings

import dill
import numpy as np

from ._parallel import (check_parallel, get_pool_for, get_rank,
                        runs_on_euler_node, runs_with_mpi)
from ._utils import capture, mute_stderr, os_write_works, with_progressbar
from .parameters import Parameters
from .persisting import _SamplerPersister, dont_persist

with mute_stderr():
    # might cause MPI related error message if run on one single node, even
    # without mpirun:
    import emcee

    emcee_version = int(emcee.__version__.split(".")[0])


RUNS_WITH_MPI = runs_with_mpi()


class _ParameterAdapter(object):
    """
    Addapts parameters argument of emcee lnprob function to parameters of uhammer
    lnprob function.
    """

    def __init__(self, names, values):
        self.names = names
        self.values = values
        # this sets every parameter as an attribute:
        self.__dict__.update(dict(zip(names, values)))

    def __str__(self):
        assignments = [
            "{}={:e}".format(name, value)
            for name, value in zip(self.names, self.values)
        ]
        return "Parameters({})".format(", ".join(assignments))


class _LnProbAdapter(object):
    """
    Adapts API of uhammer lnprob function to emcee hammer API.
    """

    def __init__(self, lnprob_uhammer, p, show_output, output_path_pattern):
        self.lnprob_uhammer = lnprob_uhammer
        self.p = p
        self.show_output = show_output
        self.output_path_pattern = output_path_pattern

    def __call__(self, parameters, *args):
        if self.output_path_pattern is not None:
            rank = get_rank()
            if rank is not None:
                output_path = self.output_path_pattern.format(worker_id=rank)
            else:
                output_path = self.output_path_pattern
        else:
            output_path = None

        with capture(output_path, self.show_output):
            return self.lnprob(parameters, args)

    def lnprob(self, parameters, args):

        p = self.p
        names = p.names

        lnprob_prior = 0
        for name, value in zip(names, parameters):
            if not p._is_in_bounds(name, value):
                return -np.inf
            lnprob_prior -= np.log(p._range_length(name))

        parameters_for_lnprob = _ParameterAdapter(names, parameters)

        computed_lnprob = self.lnprob_uhammer(parameters_for_lnprob, *args)

        if computed_lnprob > 0:
            client_function_name = self.lnprob_uhammer.__name__
            raise AssertionError(
                "{} computed positive value {} for parameters {}".format(
                    client_function_name, computed_lnprob, parameters_for_lnprob
                )
            )

        return computed_lnprob + lnprob_prior

    def __getstate__(self):
        # distributed computation requires pickling of the posterior
        # function using pickle module from the python standard library.
        # this can fail e.g. if nested functions or classes with
        # classmethods are infolved.
        # to fix that we implement here how to pickle this class using
        # dill which is more versatile than pickle from the standard
        # library:
        return dill.dumps(self.__dict__)

    def __setstate__(self, state):
        # see __getstate__
        # this method implements the unpickling part.
        self.__dict__.update(dill.loads(state))


class PickableEmceeSampler(emcee.EnsembleSampler):
    def __getstate__(self):
        dd = self.__dict__.copy()
        if "pool" in dd:
            del dd["pool"]
        return dd

    def __setstate__(self, dd):
        self.pool = None
        self.__dict__.update(dd)


def _setup_random_state(seed, sampler):
    rstate0 = sampler._random
    rstate0.seed(seed)
    random.seed(seed)
    np.random.seed(seed)


def sample(
    lnprob,
    p,
    args=None,
    n_walkers_per_param=10,
    seed=None,
    n_samples=1000,
    parallel=False,
    show_progress=False,
    show_output=True,
    output_prefix=None,
    persist=dont_persist(),
    verbose=True,
):
    """
    Runs emcee sampler to generate samples from the declared
    distribution.

    :param lnprob: Function for computing the natural logarithm of the
                   distribution to sample from. Fist argument is a
                   object with the current values of the declared
                   parameters as attributes.  The function may accept
                   additioanal arguments for auxiliary data.

    :param p:      Instance of the ``Parameters`` class.

    :param args:   List with auxiliary values. Can also be ``None`` or
                   ``[]``.

    :param n_walkers_per_param: Number of walkers per parameter.
                                Results in ``len(p) *
                                n_walkers_per_param`` walkers in total.

    :param seed:  Used random seed in case you want reproducible
                  samples. Skip this argument or use ``None`` else.

    :param n_samples: Number of samples to generate. So the return value
                      is a ``numpy`` array of shape ``(n_samples, len(p))``.

    :param parallel: Run in sequention or parallel mode? Automatically
                     detects number of cores, also detects if code is
                     run using ``mpirun``.

    :param show_progress: Show animated ASCII art to indicate progress
    of sampling.

    :param show_output: set to ``True`` if you want to see the output
                        from the workers on the terminal, else set it to
                        ``False``.

    :param output_prefix: prefix for file path to record output from the
                          workers.  Default is ``None`` which means no
                          recording of output.

    :param persist:  persisting strategy, on of the functions in
                     :py:mod:`~uhammer.persisting`.

    :param verbose:  boolean value. If set to ``True``
                     :py:func:`~uhammer.sampler.sample` prints extra information.

    This function returns two values:

    - computed samples as a ``numpy`` array of shape ``(n_samples, len(p)``
    - computed log probabilites as ``numpy`` array of shape ``(n_samples,)``.

    Samples are arranged as follows, log probbilites in similar order:

    .. table::

       +----------------------------------------+
       | samples                                |
       +========================================+
       | p-dimensional sample from walker 0     |
       +----------------------------------------+
       | p-dimensional sample from walker 1     |
       +----------------------------------------+
       | ...                                    |
       +----------------------------------------+
       | p-dimensional sample from walker n - 1 |
       +----------------------------------------+
       | p-dimensional sample from walker 0     |
       +----------------------------------------+
       | p-dimensional sample from walker 1     |
       +----------------------------------------+
       | ...                                    |
       +----------------------------------------+

    """

    if args is None:
        args = []

    assert isinstance(p, Parameters), p
    assert len(p) > 0, "need parameters"

    assert isinstance(args, (list, tuple)), args
    assert isinstance(n_walkers_per_param, int), n_walkers_per_param
    assert n_walkers_per_param > 1, "need at least 2 walkers per parameter"
    if seed is not None:
        assert isinstance(seed, int)
    assert isinstance(n_samples, int), n_samples
    assert isinstance(parallel, bool)
    assert isinstance(show_progress, bool), show_progress
    assert isinstance(verbose, bool), verbose
    assert isinstance(show_output, bool), show_output

    if output_prefix is not None:
        assert isinstance(output_prefix, str), output_prefix

    check_parallel(parallel)

    if output_prefix is not None:
        output_path_pattern = capture_output_path(output_prefix, parallel)
    else:
        output_path_pattern = None

    show_progress = _check_and_fix_progress_settings(show_output, show_progress)

    # holds paramters passed to lnprob:
    _lnprob = _LnProbAdapter(lnprob, p, show_output, output_path_pattern)
    n_walkers = n_walkers_per_param * len(p)

    # might be global including startup of workers, workers hang there:
    # workers are stopped at exit:
    pool = get_pool_for(n_walkers) if parallel else None

    if pool is not None and verbose:
        print("uhammer: started {} with {} workers".format(pool, pool.size))

    sampler = PickableEmceeSampler(n_walkers, len(p), _lnprob, args=args, pool=pool)

    sampler.n_walkers = n_walkers

    _setup_random_state(seed, sampler) if seed is not None else np.random.RandomState()
    p0 = [
        [p.start_value(name) for name in p.names] for _ in range(n_walkers)
    ]

    result = _sample(p0, sampler, n_samples, show_progress, persist, verbose=verbose)
    return result


def continue_sampling(
    persistence_file_path,
    n_samples,
    parallel=False,
    show_progress=False,
    show_output=True,
    output_prefix=None,
    persist=dont_persist(),
    verbose=True,
):
    """
    Continues sampling from a persisted sampler.

    Original arguments of  :py:func:`~uhammer.sampler.sample` like
    ``lnprob``, `n_walkers_per_param``, ``seed``, etc. are reused from
    the persisted sampler and can not be modified anymore.

    :param persistence_file_path: Path to a perviously persisted
                                  sampler. This includes `lnprob`,
                                  parameters and number of walkers.

    :param n_samples: Number of samples to generate. So the return value
                      is a ``numpy`` array of shape ``(n_samples, len(p))``.

    :param parallel: Run in sequention or parallel mode ? Automatic
                     dection of number of cores, also auto detects if
                     code is run using ``mpirun``.

    :param show_progress: Show animated ASCII art to indicate progress
                          of sampling.

    :param show_output: set to ``True`` if you want to see the output
                        from the workers on the terminal, else set it to
                        ``False``.

    :param output_prefix: prefix for file path to record output from the
                          workers.  Default is ``None`` which means no
                          recording of output.

    :param persist:  persisting strategy, on of the functions in
                     :py:mod:`~uhammer.persisting`.

    :param verbose:  boolean value, if ``True``
                     :py:func:`~uhammer.sampler.continue_sampling`
                     prints some information.

    This function returns two values:

    - computed samples as a ``numpy`` array of shape ``(n_samples, len(p)``
    - computed log probabilites as ``numpy`` array of shape ``(n_samples,)``.

    Samples are arranged as follows, log probbilites in similar order:

    .. table::

       +----------------------------------------+
       | samples                                |
       +========================================+
       | p-dimensional sample from walker 0     |
       +----------------------------------------+
       | p-dimensional sample from walker 1     |
       +----------------------------------------+
       | ...                                    |
       +----------------------------------------+
       | p-dimensional sample from walker n - 1 |
       +----------------------------------------+
       | p-dimensional sample from walker 0     |
       +----------------------------------------+
       | p-dimensional sample from walker 1     |
       +----------------------------------------+
       | ...                                    |
       +----------------------------------------+
    """
    assert isinstance(persistence_file_path, str)
    assert os.path.exists(persistence_file_path), "file {} does not exist".format(
        persistence_file_path
    )
    assert isinstance(n_samples, int), n_samples
    assert isinstance(parallel, bool)
    assert isinstance(show_progress, bool), show_progress

    assert isinstance(show_output, bool), show_output

    if output_prefix is not None:
        assert isinstance(output_prefix, str), output_prefix

    check_parallel(parallel)

    show_progress = _check_and_fix_progress_settings(show_output, show_progress)

    if output_prefix is not None:
        output_path_pattern = capture_output_path(output_prefix, parallel)
    else:
        output_path_pattern = None

    sampler = _SamplerPersister.restore_sampler(persistence_file_path)
    sampler.pool = get_pool_for(sampler.n_walkers) if parallel else None

    if emcee_version < 3:
        sampler.lnprobfn.f.show_output = show_output
        sampler.lnprobfn.f.output_path_pattern = output_path_pattern
    else:
        sampler.log_prob_fn.f.show_output = show_output
        sampler.log_prob_fn.f.output_path_pattern = output_path_pattern

    p0, lnprob0, rstate0 = sampler._last_run_mcmc_result
    return _sample(
        p0, sampler, n_samples, show_progress, persist, lnprob0, rstate0, verbose
    )


def capture_output_path(prefix, parallel):

    existing = glob.glob(prefix + "_run_*_*.txt")
    run_ids = [int(os.path.basename(p).split("_")[2]) for p in existing]
    if run_ids:
        next_run_id = max(run_ids) + 1
    else:
        next_run_id = 0

    if not parallel:
        return "{prefix}_run_{run_id:04d}.txt".format(prefix=prefix, run_id=next_run_id)

    return "{prefix}_run_{run_id:04d}_worker_{{worker_id}}.txt".format(
        prefix=prefix, run_id=next_run_id
    )


def _check_and_fix_progress_settings(show_output, show_progress):

    if show_output and show_progress:
        warnings.warn(
            "not supressing output might corrupt the progressbar. "
            "Better set show_output to False and set an appropriabe "
            "value for output_prefix."
        )

    if show_progress and not os_write_works():
        show_progress = False
        warnings.warn("disabled progressbar, can not write to fid 1")

    if show_progress and runs_on_euler_node():
        show_progress = False
        warnings.warn("disabled progressbar, will only clutter lsf output")

    return show_progress


def _sample(
    p0,
    sampler,
    n_samples,
    show_progress,
    persist,
    lnprob0=None,
    rstate0=None,
    verbose=True,
):
    print_ = functools.partial(print, file=sys.__stdout__)

    nsteps = int(np.ceil(n_samples / sampler.n_walkers))
    if verbose:
        print_("uhammer: perform {} steps of emcee sampler".format(nsteps))

    # eemcee sampler.sample
    sample_iter = sampler.sample(p0, iterations=nsteps)

    if show_progress:
        sample_iter = with_progressbar(nsteps, sample_iter)

    for step in itertools.count(1):
        try:
            state = next(sample_iter)
            if emcee_version < 3:
                state = state[:3]
            sampler._last_run_mcmc_result = state
        except StopIteration:
            break
        except Exception as e:
            persisted = persist(sampler, step, nsteps, e)
            if verbose and persisted:
                print_("uhammer: persist sampler due to exception")
            raise e
        persisted = persist(sampler, step, nsteps, None)
        if verbose and persisted:
            print_("uhammer: persisted sampler after iteration {}".format(step))

    n_param = sampler.chain.shape[2]
    samples = sampler.chain.swapaxes(0, 1).reshape(-1, n_param)[:n_samples, :]
    lnprobs = sampler.lnprobability.swapaxes(0, 1)[:n_samples].flatten()
    return samples, lnprobs



from __future__ import print_function

import os

import numpy as np

from uhammer import Parameters, persist_every_n_iterations, sample

sigma = .5

OUTPUT_PREFIX = os.environ.get("OUTPUT_PREFIX")
PERSIST_PATH = os.environ.get("PERSIST_PATH")


def gen_data():
    a0 = .5
    b0 = .5
    c0 = 1

    x = np.linspace(-2, 2, 100)
    y_measured = a0 + b0 * x + c0 * x ** 2 + sigma * np.random.randn(*x.shape)
    return x, y_measured


p = Parameters()
p.add("a", (0, 1))
p.add("b", (0, 1))
p.add("c", (0, 2))


def lnprob(p, x, y_measured, sigma=sigma):
    import numpy as np  # noqa
    from mpi4py import MPI  # noqa

    rank = MPI.COMM_WORLD.rank
    print(rank)
    y = p.a + p.b * x + p.c * x ** 2
    diff = (y - y_measured) / sigma
    return -np.dot(diff.T, diff) / 2


n_samples = 1500
n_walkers_per_param = 200


def main():
    samples, __ = sample(
        lnprob,
        p,
        args=gen_data(),
        n_walkers_per_param=n_walkers_per_param,
        n_samples=n_samples,
        show_progress=False,
        show_output=False,
        output_prefix=OUTPUT_PREFIX,
        parallel=True,
        verbose=True,
        persist=persist_every_n_iterations(1, PERSIST_PATH),
    )
    return samples


print("++++ START")
try:
    samples = main()
    print("++++ SAMPLES", samples.shape)
except Exception as e:
    print("+++++ ERROR")
    print(e)

print("++++ DONE")



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import numpy as np

from ._utils import check_range


def _rand_in_range(range_):
    min_, max_ = range_
    return np.random.random() * (max_ - min_) + min_


class Parameters(object):
    """
    Collection of used parameters incl. names and ranges.
    """

    def __init__(self):
        self.limits = {}
        self.start_values = {}
        self.names = []

    def add(self, name, range_, start_value=None):
        """
        Adds a parameter with given `name` and value `range_` to the collection.

        :param name: Name of type ``str``.
        :param range_: Allowed range as ``tuple`` or ``list`` of two ``float``
                       values. First value must be smaller than the second.
        :param start_value: optional value to start sampling from, reduces burn-in
                            phase if estimation for parameter exists.
        """

        check_range(range_)

        if name in self.names:
            raise ValueError("parameter named {!r} already added".format(name))

        self.limits[name] = range_
        self.start_values[name] = start_value
        self.names.append(name)

    def _is_in_bounds(self, name, value):
        min_, max_ = self.limits[name]
        return min_ <= value <= max_

    def _range_length(self, name):
        min_, max_ = self.limits[name]
        return max_ - min_

    def __len__(self):
        return len(self.names)

    def start_value(self, name):
        assert name in self.start_values, (name, self.start_values)
        given_start_value = self.start_values[name]
        if given_start_value is not None:
            return given_start_value
        return _rand_in_range(self.limits[name])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import os

import emcee
import numpy as np
import pytest

from uhammer import Parameters, distribution, gaussian, model, sample

emcee_version = emcee.__version__.split(".")[0]


@pytest.mark.parametrize("wrong_range", [0, (), (1, 2, 3), ("a", 1), (2, 1), (1, 1)])
def test_param(regtest, wrong_range):

    p = Parameters()
    with pytest.raises(ValueError) as e:
        p.add("a", wrong_range)

    print(e.value, file=regtest)


def test_double_param(regtest):
    p = Parameters()
    p.add("a", (1, 2))
    with pytest.raises(ValueError) as e:
        p.add("a", (1, 2))
    print(e.value, file=regtest)


def test_invalid_prob_density(regtest):

    p = Parameters()
    p.add("a", (0, 1))

    def lnprob(p):
        return 1

    with pytest.raises(AssertionError) as e:
        sample(
            lnprob,
            p,
            args=None,
            n_walkers_per_param=10,
            seed=42,
            n_samples=100,
            show_progress=False,
        )

    print(e.value, file=regtest)


def test_empty_parameters(regtest):

    p = Parameters()

    def lnprob(p):
        return 1

    with pytest.raises(AssertionError) as e:
        sample(
            lnprob,
            p,
            args=None,
            n_walkers_per_param=10,
            seed=42,
            n_samples=100,
            show_progress=False,
        )

    print(e.value, file=regtest)


def test_sample_from_distribution(regtest):

    regtest.identifier = "emcee_" + emcee_version

    seed = 42

    lnprob, p = distribution(gaussian(5), 1, (0, 3))

    samples, lnprobs = sample(
        lnprob,
        p,
        args=None,
        n_walkers_per_param=10,
        seed=seed,
        n_samples=1000,
        show_progress=False,
    )

    print("min:", np.min(samples), file=regtest)
    print("max:", np.max(samples), file=regtest)

    assert samples.shape == (1000, 1)
    assert lnprobs.shape == (1000,)

    # dimensions do not fit, should raise AssertionError:
    icov = np.array(((1, 0.2), (0.2, 1)))
    lnprob, p = distribution(gaussian(icov), 1, (0, 3))

    with pytest.raises(AssertionError) as e:
        samples, __ = sample(
            lnprob,
            p,
            args=None,
            n_walkers_per_param=10,
            seed=seed,
            n_samples=1000,
            show_progress=False,
        )

    print(e.value, file=regtest)

    np.random.seed(seed)
    N = 5

    cov = np.random.random((N, N))
    cov = np.dot(cov.T, cov)
    cov += 5 * np.eye(N)

    lnprob, p = distribution(gaussian(cov), N, (0, 2))

    samples, __ = sample(
        lnprob,
        p,
        args=None,
        n_walkers_per_param=10,
        seed=seed,
        n_samples=5000,
        show_progress=False,
    )

    assert samples.shape == (5000, N)
    print(samples[1000:, :].mean(axis=0), file=regtest)

    lnprob, p = distribution(gaussian(cov), N, N * [(-1, 1)])

    samples, __ = sample(
        lnprob,
        p,
        args=None,
        n_walkers_per_param=10,
        seed=seed,
        n_samples=5000,
        show_progress=False,
    )

    assert samples.shape == (5000, N)
    print(samples[1000:, :].mean(axis=0), file=regtest)


def test_line_fit(regtest, tmpdir):

    regtest.identifier = "emcee_" + emcee_version

    np.random.seed(42)

    a0 = 0.5
    b0 = 0.5
    c0 = 0.1
    sigma = 0.5

    x = np.linspace(-1, 1, 100)
    y_measured = a0 + b0 * x + c0 * x ** 2 + sigma * np.random.randn(*x.shape)

    p = Parameters()
    p.add("a", (-1, 1))
    p.add("b", (-1, 1))
    p.add("c", (-1, 1))

    def line(p, x):
        print(p)
        y = p.a + p.b * x + p.c * x ** 2
        return y

    def lnprob(p, x, y_measured):
        y = line(p, x)
        diff = y - y_measured
        return -np.dot(diff.T, diff) / sigma / 2.0

    n_samples = 3000
    n_walkers_per_param = 10

    log_file_prefix = tmpdir.join("out").strpath

    samples, lnprobs = sample(
        lnprob,
        p,
        args=[x, y_measured],
        n_walkers_per_param=n_walkers_per_param,
        seed=42,
        n_samples=n_samples,
        show_progress=False,
        show_output=False,
        output_prefix=log_file_prefix,
    )

    assert samples.shape == (n_samples, len(p))
    assert lnprobs.shape == (n_samples,)

    log_file = log_file_prefix + "_run_0000.txt"
    assert os.path.exists(log_file), log_file

    lines = open(log_file, "r").readlines()
    assert 0 < len(lines) <= n_samples
    assert all("Parameters" in line for line in lines)

    is_ = samples[::100, :]

    lnprob_via_model = model(line, gaussian(sigma), x, y_measured)

    samples, __ = sample(
        lnprob_via_model,
        p,
        args=[x, y_measured],
        n_walkers_per_param=n_walkers_per_param,
        n_samples=n_samples,
        show_progress=True,
        seed=42,
    )
    assert np.all(is_ == samples[::100, :])

    print(samples[::100], file=regtest)


def test_parameter_with_start_value():
    p = Parameters()
    p.add("a", (1, 2), 1.5)

    def lnprob(p):
        return -1

    samples, lnprobs = sample(
        lnprob,
        p,
        args=None,
        n_walkers_per_param=10,
        seed=42,
        n_samples=10,
        show_progress=False,
    )

    assert samples.shape == (10, 1)
    assert lnprobs.shape == (10,)
    assert set(samples.flatten()) == set((1.5,))
    assert set(lnprobs) == set((-1,))



#! /usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import, division, print_function

import zlib

import dill
import emcee

emcee_version = int(emcee.__version__.split(".")[0])

# Copyright © 2018 Uwe Schmitt <uwe.schmitt@id.ethz.ch>


class _SamplerPersister(object):
    def __init__(self, check, path, persist_on_exceptions):
        self.check = check
        self.path = path
        self.persist_on_exceptions = persist_on_exceptions
        self._iter_track = []

    def __call__(self, sampler, i, n_iter, exception):
        do_persist = self.check(i, n_iter)
        do_persist = do_persist or (
            exception is not None and self.persist_on_exceptions
        )
        if do_persist:
            with open(self.path, "wb") as fh:
                fh.write(zlib.compress(dill.dumps(sampler)))
            self._iter_track.append(i)
        if exception:
            raise exception
        return do_persist

    @staticmethod
    def restore_sampler(file_path):
        with open(file_path, "rb") as fh:
            return dill.loads(zlib.decompress(fh.read()))


def load_samples(path):
    """loads samples for persisted sampler file"""
    sampler = _SamplerPersister.restore_sampler(path)
    if emcee_version == 3:
        return sampler.flatchain
    n_samples = sampler.iterations * sampler.n_walkers
    return sampler.flatchain[:n_samples]


def persist_every_n_iterations(n, path):
    """Return value of this function can be passed as argument `persist` to ``sample``
    function to store state during sampling. This allows later continuation of sampling,
    e.g. after crashes.

    This persister saves state every ``n`` iterations and in case of failure.

    :param n:  persist after every ``n`` iterations of the sampler
    :param path: file path to store the relevant information.
    """

    def check(i, n_iter):
        return i > 0 and i % n == 0

    return _SamplerPersister(check, path, True)


def persist_final(path):
    """Return value of this function can be passed as argument `persist` to ``sample``
    function to store state during sampling. This allows later continuation of sampling,
    e.g. after crashes.

    This persister only stores the information when sampling finishes and in case
    of failure.

    :param path: file path to store the relevant information.
    """

    def check(i, n_iter):
        return i == n_iter

    return _SamplerPersister(check, path, True)


def persist_on_error(path):
    """Return value of this function can be passed as argument `persist` to ``sample``
    function to store state during sampling. This allows later continuation of sampling,
    e.g. after crashes.

    This persister only stores the information only in case of failure.

    :param path: file path to store the relevant information.
    """

    def check(i, n_iter):
        return False

    return _SamplerPersister(check, path, True)


def dont_persist():
    def check(i, n_iter):
        return False

    return _SamplerPersister(check, None, False)



#!/usr/bin/env python
# encoding: utf-8

from sys import version_info

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["numpy", "emcee", "dill", "schwimmbad"]

if version_info < (3, 7):
    requirements.extend(["backport_ipaddress", "backports.shutil_get_terminal_size"])

setup(
    version="0.3.1",  # no need to change version number in other places.
    author="Uwe Schmitt",
    author_email="uwe.schmitt@id.ethz.ch",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Convenience layer for emcee sampler",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    name="uhammer",
    packages=find_packages(include=["uhammer"]),
    zip_safe=False,
)
